"""
``aif360.sklearn`` implements a number of fairness metrics for group fairness
and individual fairness. For guidance on which metric to use for a given
application, see our
`Guidance <http://aif360.mybluemix.net/resources#guidance>`_ page.
"""
from aif360.sklearn.metrics.metrics import *
