from aif360.datasets.dataset import Dataset
from aif360.datasets.structured_dataset import StructuredDataset
from aif360.datasets.binary_label_dataset import BinaryLabelDataset
from aif360.datasets.standard_dataset import StandardDataset
from aif360.datasets.adult_dataset import AdultDataset
from aif360.datasets.bank_dataset import BankDataset
from aif360.datasets.compas_dataset import CompasDataset
from aif360.datasets.german_dataset import GermanDataset
from aif360.datasets.meps_dataset_panel19_fy2015 import MEPSDataset19
from aif360.datasets.meps_dataset_panel20_fy2015 import MEPSDataset20
from aif360.datasets.meps_dataset_panel21_fy2016 import MEPSDataset21
from aif360.datasets.regression_dataset import RegressionDataset
from aif360.datasets.law_school_gpa_dataset import LawSchoolGPADataset 
